#!/usr/bin/env python3
################################################################################
# Project:  GetOSM <https://github.com/HuidaeCho/getosm>
# Authors:  Huidae Cho
# Since:    July 11, 2021
#
# Copyright (C) 2021 Huidae Cho <https://idea.isnew.info/>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
################################################################################
"""
This script implements a wxPython demo GUI using GetOSM.
"""

import io
import sys
import wx.lib.statbmp
import threading
import queue

# https://stackoverflow.com/a/49480246/16079666
if __package__:
    from .getosm import OpenStreetMap
else:
    from getosm import OpenStreetMap


def main():
    zoomer = None
    zoomer_queue = queue.Queue()
    dzoom = 0.1

    lat = 0
    lon = 0
    zoom = 0

    def on_mouse(event):
        def zoom(x, y, dz, cancel_event):
            if not cancel_event.wait(0.01) and osm.redownload():
                zoomer_queue.put(osm.draw)

        def check_zoomer():
            nonlocal zoomer

            try:
                draw_map = zoomer_queue.get_nowait()
            except:
                zoomer.checker = wx.CallLater(0, check_zoomer)
            else:
                draw_map()

        nonlocal zoomer

        if event.ButtonDown(wx.MOUSE_BTN_LEFT):
            osm.grab(event.x, event.y)
        elif event.Dragging():
            osm.drag(event.x, event.y)
        elif event.WheelDelta > 0:
            if zoomer:
                zoomer.cancel_event.set()
                osm.cancel = True
                zoomer.join()
                osm.cancel = False
                zoomer.checker.Stop()

                cancel_event = zoomer.cancel_event
                cancel_event.clear()
            else:
                cancel_event = threading.Event()

            dz = event.WheelRotation / event.WheelDelta * dzoom

            # if used without osm.draw(), it works; otherwise, only osm.draw()
            # is visible; timing?
            osm.rescale(event.x, event.y, dz)
            zoomer = threading.Thread(target=zoom, args=(event.x, event.y, dz,
                                                         cancel_event))
            zoomer.cancel_event = cancel_event
            zoomer.checker = wx.CallLater(0, check_zoomer)
            zoomer.start()

    app = wx.App()
    root = wx.Frame(None, title="GetOSM wxPython Demo GUI", size=(800, 800))

    map_canvas = wx.lib.statbmp.GenStaticBitmap(root, wx.ID_ANY, wx.NullBitmap,
                                                size=root.Size)
    map_canvas.Bind(wx.EVT_MOUSE_EVENTS, on_mouse)
    map_canvas.Bind(wx.EVT_SIZE, lambda e: osm.resize(e.Size.Width,
                                                          e.Size.Height))

    osm = OpenStreetMap(
            wx.Image,
            lambda image: map_canvas.SetBitmap(wx.Bitmap(image)),
            lambda data: wx.Image(io.BytesIO(data)),
            lambda image, tile, x, y: image.Paste(tile, x, y),
            lambda tile, dz: tile.Scale(tile.Width*2**dz, tile.Height*2**dz),
            map_canvas.Size.Width, map_canvas.Size.Height,
            lat, lon, zoom)

    root.Show()
    app.MainLoop()


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python3
################################################################################
# Project:  GetOSM <https://github.com/HuidaeCho/getosm>
# Authors:  Huidae Cho
# Since:    July 11, 2021
#
# Copyright (C) 2021 Huidae Cho <https://idea.isnew.info/>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
################################################################################
"""
This script implements a tkinter demo GUI using GetOSM.
"""

import sys
import tkinter as tk
from tkinter import ttk
import threading
import queue
import textwrap
import webbrowser

# https://stackoverflow.com/a/49480246/16079666
if __package__:
    from .getosm import OpenStreetMap
else:
    from getosm import OpenStreetMap


def main():
    tag_map = "map"
    tag_geoms = "geoms"
    tag_github = "github"
    github_url = "https://github.com/HuidaeCho/getosm"
    zoomer = None
    zoomer_queue = queue.Queue()
    dzoom = 0.1
    dragged = False
    drawing_bbox = False
    complete_drawing = False
    prev_xy = []
    curr_geom = []
    geoms = []

    lat = 0
    lon = 0
    zoom = 0

    def draw_map(x, y):
        osm.draw()
        draw_geoms(x, y)

    def draw_geoms(x=None, y=None):
        point_size = 4
        point_half_size = point_size // 2
        outline = "blue"
        width = 2
        fill = "blue"
        stipple = "gray12"

        map_canvas.delete(tag_geoms)

        if curr_geom and x and y:
            latlon = list(osm.canvas_to_latlon(x, y))

            all_geoms = geoms.copy()
            g = curr_geom.copy()
            g.append(latlon)

            if drawing_bbox:
                ng = len(g)
                if ng > 0:
                    s = g[ng-1][0]
                    n = g[0][0]
                    w = g[0][1]
                    e = g[ng-1][1]
                    all_geoms.extend(["bbox", [s, n, w, e]])
            elif g:
                if prev_xy:
                    latlon[1] = adjust_lon(prev_xy[0], x,
                                           curr_geom[len(curr_geom)-1][1],
                                           latlon[1])
                g.append(latlon)
                all_geoms.extend(["poly", g])
        else:
            all_geoms = geoms.copy()

        geom_type = "point"
        g = 0
        ngeoms = len(all_geoms)
        while g < ngeoms:
            geom = all_geoms[g]
            if geom in ("point", "poly", "bbox"):
                geom_type = geom
                g += 1
                geom = all_geoms[g]
            if type(geom) == list:
                if geom_type == "point":
                    for xy in osm.get_xy([geom]):
                        x, y = xy[0]
                        oval = (x - point_half_size, y - point_half_size,
                                x + point_half_size, y + point_half_size)
                        map_canvas.create_oval(oval, outline=outline,
                                               width=width, fill=fill,
                                               tag=tag_geoms)
                elif geom_type == "poly":
                    for xy in osm.get_xy(geom):
                        map_canvas.create_polygon(xy, outline=outline,
                                                  width=width, fill=fill,
                                                  stipple=stipple,
                                                  tag=tag_geoms)
                else:
                    for xy in osm.get_bbox_xy(geom):
                        map_canvas.create_rectangle(xy, outline=outline,
                                                    width=width, fill=fill,
                                                    stipple=stipple,
                                                    tag=tag_geoms)
            g += 1

    def adjust_lon(prev_x, x, prev_lon, lon):
        dlon = lon - prev_lon
        if x - prev_x > 0:
            if dlon < 0:
                lon += 360
            elif dlon > 360:
                lon -= 360
        elif dlon > 0:
            lon -= 360
        elif dlon < -360:
            lon += 360
        return lon

    def zoom_map(x, y, dz):
        def zoom(x, y, dz, cancel_event):
            if not cancel_event.wait(0.01) and osm.redownload():
                zoomer_queue.put((draw_map, x, y))

        def check_zoomer():
            nonlocal zoomer

            try:
                draw_map, x, y = zoomer_queue.get_nowait()
            except:
                zoomer.checker = map_canvas.after_idle(check_zoomer)
            else:
                draw_map(x, y)

        nonlocal zoomer

        if zoomer:
            zoomer.cancel_event.set()
            osm.cancel = True
            zoomer.join()
            osm.cancel = False
            map_canvas.after_cancel(zoomer.checker)

            cancel_event = zoomer.cancel_event
            cancel_event.clear()
        else:
            cancel_event = threading.Event()

        osm.rescale(x, y, dz)
        zoomer = threading.Thread(target=zoom, args=(x, y, dz, cancel_event))
        zoomer.cancel_event = cancel_event
        zoomer.checker = map_canvas.after_idle(check_zoomer)
        zoomer.start()

    def on_drag(event):
        nonlocal dragged

        osm.drag(event.x, event.y)
        draw_geoms(event.x, event.y)
        dragged = True

    def on_move(event):
        latlon = osm.canvas_to_latlon(event.x, event.y)
        coor_label.config(text=f" {latlon[0]:.4f}, {latlon[1]:.4f} ")
        draw_geoms(event.x, event.y)

    def on_draw(event):
        nonlocal dragged, drawing_bbox, complete_drawing

        if complete_drawing:
            geoms_string = ""
            geom = []
            if drawing_bbox:
                if len(curr_geom) == 2:
                    s = curr_geom[1][0]
                    n = curr_geom[0][0]
                    w = curr_geom[0][1]
                    e = curr_geom[1][1]
                    geom.extend(["bbox", [s, n, w, e]])
                    geoms_string = f"bbox {s:.4f},{n:.4f},{w:.4f},{e:.4f}"
                drawing_bbox = False
            elif len(curr_geom) == 1:
                lat, lon = curr_geom[0]
                geom.extend(["point", [lat, lon]])
                geoms_string = f"point {lat:.4f},{lon:.4f}"
            elif curr_geom:
                geom.extend(["poly", curr_geom.copy()])
                geoms_string = "poly"
                for g in curr_geom:
                    lat, lon = g
                    geoms_string += f" {lat:.4f},{lon:.4f}"
            geoms.extend(geom)
            curr_geom.clear()
            prev_xy.clear()
            if geoms_string:
                geoms_string += "\n"
                # https://stackoverflow.com/a/35855352/16079666
                # don't use .selection_get()
                ranges = geoms_text.tag_ranges(tk.SEL)
                if ranges:
                    index = ranges[0].string
                else:
                    index = geoms_text.index(tk.INSERT)
                line, col = list(map(lambda x: int(x), index.split(".")))
                if col > 0:
                    geoms_string = "\n" + geoms_string
                    line += 1
                if ranges:
                    geoms_text.replace(*ranges, geoms_string)
                else:
                    geoms_text.insert(tk.INSERT, geoms_string)
                geoms_text.mark_set(tk.INSERT, f"{line+1}.0")
                notebook.select(geoms_frame)
                draw_geoms()
        elif not dragged:
            # https://anzeljg.github.io/rin2/book2/2405/docs/tkinter/event-handlers.html
            if event.state & 0x4:
                # Control + ButtonRelease-1
                drawing_bbox = True
                curr_geom.clear()
            elif drawing_bbox and len(curr_geom) == 2:
                del curr_geom[1]
            latlon = list(osm.canvas_to_latlon(event.x, event.y))
            if not drawing_bbox:
                if prev_xy:
                    latlon[1] = adjust_lon(prev_xy[0], event.x,
                                           curr_geom[len(curr_geom)-1][1],
                                           latlon[1])
                prev_xy.clear()
                prev_xy.extend([event.x, event.y])
            curr_geom.append(latlon)

        dragged = False
        complete_drawing = False

    def on_complete_drawing(event):
        nonlocal complete_drawing

        # XXX: sometimes, double-click events occur for both clicks and there
        # is no reliable way to register the first click only using
        # complete_drawing; a hacky way to handle such cases
        if not curr_geom:
            curr_geom.append(osm.canvas_to_latlon(event.x, event.y))
            prev_xy = [event.x, event.y]
        complete_drawing = True

    def on_cancel_drawing(event):
        nonlocal drawing_bbox

        drawing_bbox = False
        curr_geom.clear()
        prev_xy.clear()
        draw_geoms()

    def on_clear_drawing(event):
        geoms_text.delete("1.0", tk.END)
        geoms.clear()

    def ok():
        root.destroy()
        print(geoms)

    #####
    # GUI

    # root window
    root = tk.Tk()
    root_width = 800
    root_height = root_width
    root.geometry(f"{root_width}x{root_height}")
    root.resizable(False, False)
    root.title("GetOSM tkinter Demo GUI")
    # https://stackoverflow.com/a/5871414/16079666
    root.bind_class("Text", "<Control-a>",
                    lambda e: e.widget.tag_add(tk.SEL, "1.0", tk.END))

    ###########
    # top frame
    map_canvas_width = root_width
    map_canvas_height = root_height // 2

    map_canvas = tk.Canvas(root, height=map_canvas_height)
    map_canvas.pack(fill=tk.BOTH)

    osm = OpenStreetMap(
            lambda width, height: map_canvas.delete(tag_map),
            lambda image: map_canvas.tag_lower(tag_map),
            lambda data: tk.PhotoImage(data=data),
            lambda image, tile, x, y:
                map_canvas.create_image(x, y, anchor=tk.NW, image=tile,
                                        tag=tag_map),
            lambda tile, dz: tile.zoom(2**abs(dz)) if dz > 0 else
                             tile.subsample(2**abs(dz)),
            map_canvas_width, map_canvas_height,
            lat, lon, zoom)

    map_canvas.bind("<ButtonPress-1>", lambda e: osm.grab(e.x, e.y))
    map_canvas.bind("<B1-Motion>", on_drag)
    # Linux
    # https://anzeljg.github.io/rin2/book2/2405/docs/tkinter/event-types.html
    map_canvas.bind("<Button-4>", lambda e: zoom_map(e.x, e.y, dzoom))
    map_canvas.bind("<Button-5>", lambda e: zoom_map(e.x, e.y, -dzoom))
    # Windows and macOS
    # https://anzeljg.github.io/rin2/book2/2405/docs/tkinter/event-types.html
    map_canvas.bind("<MouseWheel>",
                    lambda e: zoom_map(e.x, e.y,
                                       dzoom if e.delta > 0 else -dzoom))
    map_canvas.bind("<Motion>", on_move)
    map_canvas.bind("<ButtonRelease-1>", on_draw)
    map_canvas.bind("<Double-Button-1>", on_complete_drawing)
    map_canvas.bind("<ButtonRelease-3>", on_cancel_drawing)
    map_canvas.bind("<Double-Button-3>", on_clear_drawing)

    ####################
    # bottom frame
    notebook_width = root_width
    notebook = ttk.Notebook(root, width=notebook_width)

    geoms_frame = tk.Frame(notebook)
    notebook.add(geoms_frame, text="Geometries")

    help_frame = tk.Frame(notebook)
    notebook.add(help_frame, text="Help")

    notebook.pack(side=tk.LEFT, fill=tk.BOTH, expand=True)

    #############
    # geoms frame
    geoms_top_frame = tk.Frame(geoms_frame)
    geoms_top_frame.pack(fill=tk.BOTH, expand=True)

    # text for geoms
    geoms_text = tk.Text(geoms_top_frame, width=20, height=1, wrap=tk.NONE)
    geoms_text.bind("<Key>", lambda e: "break" if e.state == 0 else None)
    geoms_text.pack(side=tk.LEFT, fill=tk.BOTH, expand=True)

    # vertical scroll bar for geoms
    geoms_vscrollbar = tk.Scrollbar(geoms_top_frame)
    geoms_vscrollbar.config(command=geoms_text.yview)
    geoms_vscrollbar.pack(side=tk.LEFT, fill=tk.Y)
    geoms_text.config(yscrollcommand=geoms_vscrollbar.set)

    # horizontal scroll bar for geoms
    geoms_hscrollbar = tk.Scrollbar(geoms_frame, orient=tk.HORIZONTAL)
    geoms_hscrollbar.config(command=geoms_text.xview)
    geoms_hscrollbar.pack(fill=tk.X)
    geoms_text.config(xscrollcommand=geoms_hscrollbar.set)

    geoms_bottom_frame = tk.Frame(geoms_frame)
    geoms_bottom_frame.pack(fill=tk.BOTH)

    # buttons
    tk.Button(geoms_bottom_frame, text="OK", command=ok).pack(
            side=tk.LEFT, expand=True)
    tk.Button(geoms_bottom_frame, text="Cancel", command=root.destroy).pack(
            side=tk.LEFT, expand=True)

    ############
    # help frame

    # text for help
    help_text = tk.Text(help_frame, width=20, height=1, wrap=tk.NONE)
    help_text.insert(tk.END, textwrap.dedent(f"""\
            Map operations
            ==============
            Pan:                        Drag using left button
            Zoom:                       Scroll
            Draw point:                 Double left click
            Start drawing poly:         Left click
            Start drawing bbox:         Control + left click
            Complete drawing poly/bbox: Double left click
            Cancel drawing poly/bbox:   Right click
            Clear geometries:           Double right click

            GitHub repository
            =================
            {github_url}"""))
    help_text.tag_add(tag_github, "end - 1 line", "end")
    help_text.tag_config(tag_github, foreground="blue", underline=True)
    help_text.tag_bind(tag_github, "<Enter>",
                       lambda e: help_text.config(cursor="hand2"))
    help_text.tag_bind(tag_github, "<Leave>",
                       lambda e: help_text.config(cursor=""))
    help_text.tag_bind(tag_github, "<Button-1>",
                       lambda e: webbrowser.open(github_url))
    help_text.config(state=tk.DISABLED)
    help_text.pack(fill=tk.BOTH, expand=True)

    # label for coordinates
    coor_label = tk.Label(notebook)
    coor_label.place(relx=1, rely=0, anchor=tk.NE)

    #########
    # run GUI
    root.mainloop()


if __name__ == "__main__":
    sys.exit(main())



################################################################################
# Project:  GetOSM <https://github.com/HuidaeCho/getosm>
# Authors:  Huidae Cho
# Since:    July 11, 2021
#
# Copyright (C) 2021 Huidae Cho <https://idea.isnew.info/>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
################################################################################
"""
This module provides an OpenStreetMap downloader.
"""

import collections
import sys
import math
import urllib.request


class Tile:
    def __init__(self, key, x, y, z):
        self.key = key
        self.x = x
        self.y = y
        self.z = z
        self.dz = 0
        self.rescaled_image = None


class CachedTile:
    def __init__(self, image, raw):
        self.image = image
        self.raw = raw


class OpenStreetMap:
    def __init__(self, create_image, draw_image, create_tile, draw_tile,
                 resample_tile, width=256, height=256, lat=0, lon=0, z=0,
                 verbose=False):
        self.create_image = create_image
        self.draw_image = draw_image
        self.create_tile = create_tile
        self.draw_tile = draw_tile
        self.resample_tile = resample_tile
        self.width = width
        self.height = height
        self.lat = lat
        self.lon = lon
        self.z = z
        self.verbose = verbose

        self.z_min = 0
        self.z_max = 18
        self.lat_min = -85.0511
        self.lat_max = 85.0511
        self.dz = 0
        self.tiles = []
        self.rescaled_tiles = []
        # TODO: Tile caching mechanism
        self.cached_tiles = {}
        self.cancel = False

        self.redownload()
        self.draw()

    def message(self, *args, end=None):
        if self.verbose:
            print(*args, end=end, file=sys.stderr, flush=True)

    def resize(self, width, height):
        self.width = width
        self.height = height
        self.max_cached_tiles = int(2 * (width / 256) * (height / 256))
        self.redownload()

    # Adapted from https://stackoverflow.com/a/62607111/16079666
    # https://wiki.openstreetmap.org/wiki/Slippy_map_tilenames
    # x from 0 at lon=-180 to 2**z-1 at lon=180
    # y from 0 at lat=85.0511 to 2**z-1 at lat=-85.0511
    # tilex, tiley = int(x), int(y)
    def latlon_to_tile(self, lat, lon, z):
        lat = max(min(lat, self.lat_max), self.lat_min)
        lat = math.radians(lat)
        n = 2**z
        x = (lon+180)/360*n
        y = (1-math.log(math.tan(lat)+(1/math.cos(lat)))/math.pi)/2*n
        return x, y

    def tile_to_latlon(self, x, y, z):
        n = 2**z
        lat = math.degrees(math.atan(math.sinh(math.pi*(1-2*y/n))))
        lon = x/n*360-180
        return lat, lon

    def latlon_to_canvas(self, lat, lon):
        x, y = self.latlon_to_tile(lat, lon, self.z)
        x = self.xoff + (x - self.x) * 256
        y = self.yoff + (y - self.y) * 256
        return x, y

    def canvas_to_latlon(self, x, y):
        x = self.x + (x - self.xoff) / 256
        y = self.y + (y - self.yoff) / 256
        lat, lon = self.tile_to_latlon(x, y, self.z)
        while lon < -180:
            lon += 360
        while lon > 180:
            lon -= 360
        return lat, lon

    def get_tile_url(self, x, y, z):
        return f"http://a.tile.openstreetmap.org/{z}/{x}/{y}.png"

    def download_tile(self, x, y, z):
        tile_url = self.get_tile_url(x, y, z)
        tile_key = f"{z}/{x}/{y}"
        if tile_key not in self.cached_tiles:
            # need this header to successfully download tiles from the server
            req = urllib.request.Request(tile_url, headers={
                "User-Agent": "urllib.request"
            })
            try:
                with urllib.request.urlopen(req) as f:
                    self.cached_tiles[tile_key] = CachedTile(f.read(), True)
                    self.message(f"{tile_url} downloaded")
            except Exception as e:
                self.message(f"{tile_url}: Failed to download")
        return tile_key

    def download(self, lat, lon, z):
        z = min(max(z, self.z_min), self.z_max)
        ntiles = 2**z

        # do we need it?
        # if ntiles * 256 < self.height:
        #     lat = 0

        # calculate x,y offsets to lat,lon within width,height
        xc, yc = self.latlon_to_tile(lat, lon, z)
        x, y = int(xc), int(yc)
        n, w = self.tile_to_latlon(x, y, z)
        s, e = self.tile_to_latlon(x + 1, y + 1, z)
        xo, yo = self.latlon_to_tile(n, w, z)

        xoff = int(self.width / 2 - (xc - xo) * 256)
        yoff = int(self.height / 2 - (yc - yo) * 256)

        xmin = x - math.ceil(xoff / 256)
        ymin = max(y - math.ceil(yoff / 256), 0)
        xmax = x + math.ceil((self.width - xoff - 256) / 256)
        ymax = min(y + math.ceil((self.height - yoff - 256) / 256), ntiles - 1)

        self.lat = lat
        self.lon = lon
        self.x = x
        self.y = y
        self.z = z
        self.ntiles = ntiles
        self.xmin = xmin
        self.xmax = xmax
        self.ymin = ymin
        self.ymax = ymax
        self.xoff = xoff
        self.yoff = yoff

        self.message("image size:", self.width, self.height)
        self.tiles.clear()

        for xi in range(xmin, xmax + 1):
            xt = xi % ntiles
            for yi in range(ymin, ymax + 1):
                if self.cancel:
                    self.message("download_map cancelled")
                    break
                tile_url = self.get_tile_url(xt, yi, z)
                tile_key = self.download_tile(xt, yi, z)
                tile_x = xoff + (xi - x) * 256
                while tile_x <= -256:
                    tile_x += 256 * ntiles
                while tile_x > self.width:
                    tile_x -= 256 * ntiles
                tile_y = yoff + (yi - y) * 256
                self.tiles.append(Tile(tile_key, tile_x, tile_y, z))
            if self.cancel:
                break
        return not self.cancel

    def redownload(self):
        return self.download(self.lat, self.lon, self.z)

    def draw(self):
        image = self.create_image(self.width, self.height)
        for tile in self.tiles:
            if tile.key in self.cached_tiles:
                cached_tile = self.cached_tiles[tile.key]
                if cached_tile.raw:
                    cached_tile.image = self.create_tile(cached_tile.image)
                    cached_tile.raw = False
                self.draw_tile(image, cached_tile.image, tile.x, tile.y)
        self.draw_image(image)
        # try to not clear self.rescaled_tiles to keep rescale() safer
        self.rescaled_tiles = self.tiles.copy()

    def draw_rescaled(self):
        image = self.create_image(self.width, self.height)
        for tile in self.rescaled_tiles:
            if tile.rescaled_image:
                tile_image = tile.rescaled_image
            else:
                cached_tile = self.cached_tiles[tile.key]
                if cached_tile.raw:
                    cached_tile.image = self.create_tile(cached_tile.image)
                    cached_tile.raw = False
                tile_image = cached_tile.image

            tile.rescaled_image = self.resample_tile(tile_image, tile.dz)
            self.draw_tile(image, tile.rescaled_image, tile.x, tile.y)
        self.draw_image(image)

    def grab(self, x, y):
        self.grab_x = x
        self.grab_y = y

    def drag(self, x, y, draw=True):
        dx = x - self.grab_x
        dy = y - self.grab_y
        self.grab(x, y)
        x = self.width / 2 - dx
        y = self.height / 2 - dy
        lat, lon = self.canvas_to_latlon(x, y)
        old_lat = self.lat
        self.download(lat, lon, self.z)
        if abs(old_lat - self.lat) <= sys.float_info.epsilon:
            dy = 0
        if draw:
            self.draw()
        return dx, dy

    def reset_zoom(self):
        self.dz = 0

    def zoom(self, x, y, dz, draw=True):
        zoomed = False
        self.dz += dz
        if ((self.z < self.z_max and self.dz >= 1) or
            (self.z > self.z_min and self.dz <= -1)):
            dz = 1 if self.dz > 0 else -1
            z = self.z + dz
            # pinned zoom at x,y
            if dz > 0:
                # each zoom-in doubles
                xc = (x + self.width / 2) / 2
                yc = (y + self.height / 2) / 2
                self.message("zoom in:", z)
            else:
                # each zoom-out halves
                xc = self.width - x
                yc = self.height - y
                self.message("zoom out:", z)
            # lat,lon at xc,yc
            lat, lon = self.canvas_to_latlon(xc, yc)
            zoomed = True
        elif ((self.z == self.z_max and self.dz >= 1) or
              (self.z == self.z_min and self.dz <= -1)):
            # need to download map for z_max or z_min because when the first
            # event of either zoom level was canceled, there are no cached
            # tiles
            lat, lon, z = self.lat, self.lon, self.z
            zoomed = True

        if zoomed:
            self.download(lat, lon, z)
            self.reset_zoom()
            if draw:
                self.draw()
        return zoomed

    def zoom_to_bbox(self, bbox, draw=True):
        s, n, w, e = bbox

        lat = (s + n) / 2

        if w == e or (w == -180 and e == 180):
            dlon = 360
            lon = e - 180 if e >= 0 else e
        elif w < e:
            dlon = e - w
            lon = (w + e) / 2
        else:
            dlon = 360 - w + e
            lon = (w + e) / 2
            lon = lon - 180 if lon >= 0 else lon + 180
        e = w + dlon

        # find the center
        l, t = self.latlon_to_tile(n, w, self.z)
        r, b = self.latlon_to_tile(s, e, self.z)
        lat, lon = self.tile_to_latlon((l + r) / 2, (t + b) / 2, self.z)

        # estimate z
        z_lat = math.log2((self.lat_max - self.lat_min) / (n - s))
        z_lon = math.log2(360 / dlon)
        z = math.ceil(min(z_lat, z_lon))

        # check if z covers the entire bbox
        l, t = self.latlon_to_tile(n, w, z)
        r, b = self.latlon_to_tile(s, e, z)
        width = (r - l) * 256
        height = (b - t) * 256

        if 2 * width <= self.width and 2 * height <= self.height:
            # if z is too loose, tighten it
            z += 1
        elif width > self.width or height > self.height:
            # if z is too tight, loosen it
            z -= 1

        self.download(lat, lon, z)
        if draw:
            self.draw()

        return [s, n, w, e]

    def rescale(self, x, y, dz, draw=True):
        rescaled = False
        self.dz += dz
        if ((self.z < self.z_max and self.dz >= 1) or
            (self.z > self.z_min and self.dz <= -1)):
            dz = 1 if self.dz > 0 else -1
            z = self.z + dz
            xc, yc = self.width / 2, self.height / 2
            for i in range(0, abs(dz)):
                if dz > 0:
                    # each zoom-in doubles
                    xc = (x + xc) / 2
                    yc = (y + yc) / 2
                else:
                    # each zoom-out halves
                    xc = 2 * xc - x
                    yc = 2 * yc - y

            for i in reversed(range(len(self.rescaled_tiles))):
                tile = self.rescaled_tiles[i]
                if tile.key not in self.cached_tiles:
                    del self.rescaled_tiles[i]
                    continue

                tile.x = self.width / 2 - 2**dz * (xc - tile.x)
                tile.y = self.height / 2 - 2**dz * (yc - tile.y)
                tile.dz = dz
                tile_size = 2**(z - tile.z) * 256

                if (tile.x + tile_size < 0 or tile.y + tile_size < 0 or
                    tile.x >= self.width or tile.y >= self.height or
                    # rescaling too fast can raise a memory exception:
                    # _tkinter.TclError: not enough free memory for image
                    # buffer; avoid rescaling more than 2**6 = 64 times; this
                    # number is experimental
                    z > tile.z + 6):
                    del self.rescaled_tiles[i]

            lat, lon = self.canvas_to_latlon(xc, yc)
            xt, yt = self.latlon_to_tile(lat, lon, z)
            xi, yi = int(xt), int(yt)
            n, w = self.tile_to_latlon(xi, yi, z)
            s, e = self.tile_to_latlon(xi + 1, yi + 1, z)
            xo, yo = self.latlon_to_tile(n, w, z)

            self.lat = lat
            self.lon = lon
            self.x = xi
            self.y = yi
            self.z = z
            self.xoff = int(self.width / 2 - (xt - xo) * 256)
            self.yoff = int(self.height / 2 - (yt - yo) * 256)

            rescaled = True

        if rescaled:
            self.reset_zoom()
            if draw:
                self.draw_rescaled()
        return rescaled

    def repeat_xy(self, xy):
        outxy = []
        n = self.width // (256 * self.ntiles)
        for i in range(-n//2-1, n//2+2):
            dx = i * 256 * self.ntiles
            p = []
            for coor in xy:
                x, y = coor
                x += dx
                p.append([x, y])
            outxy.append(p)
        return outxy

    def get_xy(self, latlon):
        outxy = []
        if latlon:
            xy = []
            for coor in latlon:
                xy.append(self.latlon_to_canvas(*coor))
            outxy.extend(self.repeat_xy(xy))
        return outxy

    def get_bbox_xy(self, bbox):
        outxy = []
        if bbox:
            s, n, w, e = bbox
            l, t = self.latlon_to_canvas(n, w)
            r, b = self.latlon_to_canvas(s, e)
            if w > e:
                l -= 256 * self.ntiles
            xy = [[l, t], [r, b]]
            outxy.extend(self.repeat_xy(xy))
        return outxy



import setuptools

with open("README.md") as f:
    long_description = f.read().rstrip()

with open("getosm/VERSION") as f:
    version = f.read().rstrip()

setuptools.setup(
    name="getosm",
    version=version,
    license="GPLv3+",
    author="Huidae Cho",
    author_email="grass4u@gmail.com",
    description="GetOSM is an OpenStreetMap downloader written in Python that is agnostic of GUI frameworks.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/HuidaeCho/getosm",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3",
    package_data={"getosm": ["VERSION"]},
    entry_points={"console_scripts": [
        "osmtk=getosm.osmtk:main",
        "osmwx=getosm.osmwx:main"
    ]},
)
